///////////////////////////////////////////////
// Copyright
///////////////////////////////////////////////
//
// TextFX8
// Copyright (c) 1995-2018 Jari Komppa
//
//
///////////////////////////////////////////////
// License
///////////////////////////////////////////////
// 
//     This software is provided 'as-is', without any express or implied
//     warranty.    In no event will the authors be held liable for any damages
//     arising from the use of this software.
// 
//     Permission is granted to anyone to use this software for any purpose,
//     including commercial applications, and to alter it and redistribute it
//     freely, subject to the following restrictions:
// 
//     1. The origin of this software must not be misrepresented; you must not
//        claim that you wrote the original software. If you use this software
//        in a product, an acknowledgment in the product documentation would be
//        appreciated but is not required.
//     2. Altered source versions must be plainly marked as such, and must not be
//        misrepresented as being the original software.
//     3. This notice may not be removed or altered from any source distribution.
// 
// (eg. same as ZLIB license)
// 
//
///////////////////////////////////////////////

#ifndef TEXTFX_H_INCLUDED
#define TEXTFX_H_INCLUDED

// Build options
#define TFX_SLOW_GENERATORS
#define USE_STB_IMAGE // needs stb_image instantiated somewhere.

// Console parameters
static const int TFX_ConsoleHeight = 50;
static const int TFX_ConsoleWidth = 80;

// The raw text mode buffer that will be shown at the next TFX_Present()
extern unsigned short TFX_FrameBuffer[];

// Main function parameters.
// If 0, app was built as windows application, not a console one
extern int TFX_Paramc; 
extern char **TFX_Params;

extern unsigned char TFX_AsciiFontdata[12 * 256];
extern int TFX_Palette8[16 * 3];
extern int TFX_Palette6[16 * 3];
extern int TFX_PaletteIBM[16 * 3];
extern int TFX_Palette32[16];

#define IS_BIT(a,b) (((a) & (1 << (b))) ? 1 : 0)
#define SQUARE(a) ((a) * (a))
#define MAX(a,b) (((a) > (b)) ? (a) : (b))

// Present framebuffer to user
extern void TFX_Present();

// Set the title of the window
extern void TFX_SetTitle(char *aTitle);

// Print a string in TFX_FrameBuffer.
extern void TFX_printf(int x, int y, int color, char * format, ...);

//////////// Converters

// Expects a 64*64*64*64 index lookup table
extern unsigned short *lut_color2x2;
void alloc_lut_color2x2();
void load_lut_color2x2(char * aFilename);
#ifdef USE_STB_IMAGE
void load_lut_color2x2_png(char * aFilename);
#endif
void convert_color2x2_160x100(unsigned int *fb, unsigned short *dst = TFX_FrameBuffer);

// Expects a 3^16 index lookup table (some 41 million indices)
extern unsigned short *lut_hires;
void alloc_lut_hires();
void load_lut_hires(char * aFilename);
#ifdef USE_STB_IMAGE
void load_lut_hires_png(char * aFilename);
#endif
void convert_hires_320x200(unsigned int *fb, unsigned short *dst = TFX_FrameBuffer);

// Expects a 64^3 index lookup table.
extern unsigned short *lut_blockcolor;
void alloc_lut_blockcolor();
void build_lut_blockcolor();
void build_lut_longramp(); // alternate, experimental
void convert_blockcolor_80x50(unsigned int *aSource, unsigned short *aTarget = TFX_FrameBuffer);
void convert_blockcolor_160x100(unsigned int *aSource, unsigned short *aTarget = TFX_FrameBuffer);
void convert_blockcolor_320x200(unsigned int *aSource, unsigned short *aTarget = TFX_FrameBuffer);

// no lookup table!
void convert_halfblockcolor_160x100(unsigned int *aSource, unsigned short *aTarget = TFX_FrameBuffer);

// expects 16^4 char lookup table.
extern unsigned char *lut_asciiart;
void alloc_lut_asciiart();
void build_lut_asciiart();
void convert_asciiart_160x100(unsigned int *aSource, unsigned short *aTarget = TFX_FrameBuffer);

// uses blockcolor and halfblock internally
void convert_blockmixer_160x100(unsigned int *aSource, unsigned short *aTarget = TFX_FrameBuffer);


//////////// Lookup table generators

// Note: these are not for real-time use. Pre-calculated tables
// are expected. Runtime is in hours (for ~3GHz class machines)

#ifdef TFX_SLOW_GENERATORS
void extract_colors_2x2();
void build_lut_2x2();
void extract_colors_4x4();
void build_lut_4x4();
#endif

#endif // TEXTFX_H_INCLUDED
