///////////////////////////////////////////////
// Copyright
///////////////////////////////////////////////
//
// TextFX8
// Copyright (c) 1995-2018 Jari Komppa
//
//
///////////////////////////////////////////////
// License
///////////////////////////////////////////////
// 
//     This software is provided 'as-is', without any express or implied
//     warranty.    In no event will the authors be held liable for any damages
//     arising from the use of this software.
// 
//     Permission is granted to anyone to use this software for any purpose,
//     including commercial applications, and to alter it and redistribute it
//     freely, subject to the following restrictions:
// 
//     1. The origin of this software must not be misrepresented; you must not
//        claim that you wrote the original software. If you use this software
//        in a product, an acknowledgment in the product documentation would be
//        appreciated but is not required.
//     2. Altered source versions must be plainly marked as such, and must not be
//        misrepresented as being the original software.
//     3. This notice may not be removed or altered from any source distribution.
// 
// (eg. same as ZLIB license)
// 
//
///////////////////////////////////////////////

#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include "textfx.h"

#define TFX_COLMAP(r,g,b) *(lut_blockcolor + ((r)<<12) + ((g)<<6) + (b))

unsigned short *lut_blockcolor = 0;

void alloc_lut_blockcolor()
{
	lut_blockcolor = new unsigned short[64 * 64 * 64];
}

static int allowed(int color1, int color2)
{
	// disallow certain color combinations
    /*
	000 0
	001 1
	010 2
	011 3
	100 4
	101 5
	110 6
	111 7
	111 8
	002 9
	020 10
	022 11
	200 12
	202 13
	220 14
	222 15
	*/

	int mask[16] = {
	0x000,
	0x001,
	0x010,
	0x011,
	0x100,
	0x101,
	0x110,
	0x111,
	0x111,
	0x001,
	0x010,
	0x011, 
	0x100,
	0x101,
	0x110,
	0x111
	};

	if (color1 > color2)
	{
		int t = color2;
		color2 = color1;
		color1 = t;
	}

	if (color1 == 0 && color2 > 8) return 0;
	if (color1 < 7 && color2 > 8  && !(mask[color1] & mask[color2])) return 0;
	return 1;
}

static void calcColor(int blue, int green, int red, int offset) 
{
    int a, b, c, d, ch, co, R, G, B;
    int lastdist, dist;

    lastdist = 0x7fffffff;

    for (c = 0, d = 0; c < 16; c++, d += 3) 
    {
        dist = SQUARE(TFX_Palette8[d + 0] - red) +
            SQUARE(TFX_Palette8[d + 1] - green) +
            SQUARE(TFX_Palette8[d + 2] - blue);
        if (dist < lastdist) 
        {
            lastdist = dist;
            co = c;
            ch = 219; // 100% block in IBMSCII 
            R = TFX_Palette8[d + 0];
            G = TFX_Palette8[d + 1];
            B = TFX_Palette8[d + 2];
        }
    }
    c = co;
    d = c * 3;
    for (b = 0, a = 0; b < 16; b++, a += 3) 
    {
		if (allowed(a/3, d/3))
		{
			dist = SQUARE(((TFX_Palette8[a + 0] + TFX_Palette8[d + 0]) / 2) - red) +
				SQUARE(((TFX_Palette8[a + 1] + TFX_Palette8[d + 1]) / 2) - green) +
				SQUARE(((TFX_Palette8[a + 2] + TFX_Palette8[d + 2]) / 2) - blue);
			if (dist < lastdist) 
			{
				lastdist = dist;
				if (b > c)
					co = b + (c << 4);
				else
					co = c + (b << 4);
				ch = 177; // 50% block in IBMSCII 

				R = (TFX_Palette8[a + 0] + TFX_Palette8[d + 0]) / 2;
				G = (TFX_Palette8[a + 1] + TFX_Palette8[d + 1]) / 2;
				B = (TFX_Palette8[a + 2] + TFX_Palette8[d + 2]) / 2;
			}
			dist = SQUARE((TFX_Palette8[a + 0] * 2 / 3 + TFX_Palette8[d + 0] / 3) - red) +
				SQUARE((TFX_Palette8[a + 1] * 2 / 3 + TFX_Palette8[d + 1] / 3) - green) +
				SQUARE((TFX_Palette8[a + 2] * 2 / 3 + TFX_Palette8[d + 2] / 3) - blue);
			if (dist < lastdist) 
			{
				lastdist = dist;
				if (b > c)
				{
					co = b + (c << 4);
					ch = 178; // 75% block in IBMSCII 
				}
				else
				{
					co = c + (b << 4);
					ch = 176; // 25% block in IBMSCII 
				}

				R = (TFX_Palette8[a + 0] * 2 + TFX_Palette8[d + 0]) / 3;
				G = (TFX_Palette8[a + 1] * 2 + TFX_Palette8[d + 1]) / 3;
				B = (TFX_Palette8[a + 2] * 2 + TFX_Palette8[d + 2]) / 3;
			}
			dist = SQUARE((TFX_Palette8[a + 0] / 3 + TFX_Palette8[d + 0] * 2 / 3) - red) +
				SQUARE((TFX_Palette8[a + 1] / 3 + TFX_Palette8[d + 1] * 2 / 3) - green) +
				SQUARE((TFX_Palette8[a + 2] / 3 + TFX_Palette8[d + 2] * 2 / 3) - blue);
			if (dist < lastdist) 
			{
				lastdist = dist;
				if (c > b)
				{
					co = c + (b << 4);
					ch = 178; // 75% block in IBMSCII 
				}
				else
				{
					co = b + (c <<4 );
					ch = 176; // 25% block in IBMSCII 
				}

				R = (TFX_Palette8[a + 0] + TFX_Palette8[d + 0] * 2 ) / 3;
				G = (TFX_Palette8[a + 1] + TFX_Palette8[d + 1] * 2 ) / 3;
				B = (TFX_Palette8[a + 2] + TFX_Palette8[d + 2] * 2 ) / 3;
			}
		}
    }

    lut_blockcolor[offset] = (co << 8) + ch;
}


void build_lut_blockcolor() 
{
    int r, g, b;
	if (lut_blockcolor == 0) alloc_lut_blockcolor();
    for (r = 0; r < 64; r++) 
    {
        for (g = 0; g < 64; g++)
        {
            for (b = 0; b < 64; b++)
            {
                calcColor(r * 4, g * 4, b * 4, (r << 12) + (g << 6) + b);
            }
        }
    }
}

#define TFX_RAMPLENGTH 9
static unsigned char rampchars[9] = { 0x20, 0xFA, 0xF9, 0x07, 0x2b, 0x09, 0x4f, 0x0f, 0xe9 }; 

static void calcColorLongramp(int blue, int green, int red, int offset) 
{
    int ch, co, R, G, B;
    int lastdist, dist;

    lastdist = 0x7fffffff;

    int rampidx;
    int fg;
    int bg;

    for (rampidx = 0; rampidx < TFX_RAMPLENGTH; rampidx++)
    {
        for (bg = 0; bg < 16; bg++)
        {
            for (fg = 0; fg < 16; fg++)
            {                
                dist = SQUARE(((TFX_Palette8[fg*3 + 0] * rampidx + TFX_Palette8[bg*3 + 0] * (TFX_RAMPLENGTH * 2 - rampidx))) - red * TFX_RAMPLENGTH * 2) +
                       SQUARE(((TFX_Palette8[fg*3 + 1] * rampidx + TFX_Palette8[bg*3 + 1] * (TFX_RAMPLENGTH * 2 - rampidx))) - green * TFX_RAMPLENGTH * 2) +
                       SQUARE(((TFX_Palette8[fg*3 + 2] * rampidx + TFX_Palette8[bg*3 + 2] * (TFX_RAMPLENGTH * 2 - rampidx))) - blue * TFX_RAMPLENGTH * 2);

                if (dist < lastdist) 
                {
                    lastdist = dist;
                    co = (bg << 4) | fg;
                    ch = rampchars[rampidx];
                    
                    R = ((TFX_Palette8[fg*3 + 0] * rampidx + TFX_Palette8[bg*3 + 0] * (TFX_RAMPLENGTH * 2 - rampidx))) / (TFX_RAMPLENGTH * 2);
                    G = ((TFX_Palette8[fg*3 + 1] * rampidx + TFX_Palette8[bg*3 + 1] * (TFX_RAMPLENGTH * 2 - rampidx))) / (TFX_RAMPLENGTH * 2);
                    B = ((TFX_Palette8[fg*3 + 2] * rampidx + TFX_Palette8[bg*3 + 2] * (TFX_RAMPLENGTH * 2 - rampidx))) / (TFX_RAMPLENGTH * 2);
                    
                }
            }
        }
    }

    lut_blockcolor[offset] = (co << 8) + ch;
}


void build_lut_longramp() 
{
    int r, g, b;
	if (lut_blockcolor == 0) alloc_lut_blockcolor();
    for (r = 0; r < 64; r++) 
    {
        for (g = 0; g < 64; g++)
        {
            for (b = 0; b < 64; b++)
            {
                calcColorLongramp(r * 4, g * 4, b * 4, (r << 12) + (g << 6) + b);
            }
        }
    }
}



void convert_blockcolor_80x50(unsigned int *aSource, unsigned short *aTarget)
{
	int x, y, c;
    for (y = 0, c = 0; y < 50; y++)
    {
        for (x = 0; x < 80; x++, c++)
        {
            int color = aSource[c];
            aTarget[c] = TFX_COLMAP(((color) & 0xff) >> 2,
                ((color >> 8) & 0xff) >> 2,
                ((color >> 16) & 0xff) >> 2);
        }
    }
}



void convert_blockcolor_160x100(unsigned int *aSource, unsigned short *aTarget)
{
	int x, y, c;
    for (y = 0, c = 0; y < 50; y++)
    {
        for (x = 0; x < 80; x++, c++)
        {
            int redblue;
            int green;
            redblue  = aSource[(y*2+0)*160+x*2+0] & 0xff00ff;
            green    = aSource[(y*2+0)*160+x*2+0] & 0x00ff00;
            redblue += aSource[(y*2+0)*160+x*2+1] & 0xff00ff;
            green   += aSource[(y*2+0)*160+x*2+1] & 0x00ff00;
            redblue += aSource[(y*2+1)*160+x*2+0] & 0xff00ff;
            green   += aSource[(y*2+1)*160+x*2+0] & 0x00ff00;
            redblue += aSource[(y*2+1)*160+x*2+1] & 0xff00ff;
            green   += aSource[(y*2+1)*160+x*2+1] & 0x00ff00;
            int color = ((redblue>>2) & 0xff00ff) + ((green>>2) & 0x00ff00);

            int r = color & 0xff;
            int g = (color >> 8) & 0xff;
            int b = (color >> 16) & 0xff;

            int offset = ((r >> 2)<<12) + 
                         ((g >> 2)<<6) + 
                         (b >> 2);

            aTarget[c] = lut_blockcolor[offset];
        }
    }
}


void convert_blockcolor_320x200(unsigned int *aSource, unsigned short *aTarget)
{
	int x, y, c;
    for (y = 0, c = 0; y < 50; y++)
    {
        for (x = 0; x < 80; x++, c++)
        {
            int redblue = 0;
            int green = 0;
			int iny, inx;
            for (iny = 0; iny < 4; iny++)
            {
                for (inx = 0; inx < 4; inx++)
                {
                    redblue += aSource[(y*4+iny)*320+x*4+inx] & 0xff00ff;
                    green   += aSource[(y*4+iny)*320+x*4+inx] & 0x00ff00;
                }
            }

            int color = ((redblue>>4) & 0xff00ff) + ((green>>4) & 0x00ff00);

            aTarget[c] = TFX_COLMAP(
				((color>> 0) & 0xff)>>2,
                ((color>> 8) & 0xff)>>2,
                ((color>>16) & 0xff)>>2);
        }
    }
}
