///////////////////////////////////////////////
// Copyright
///////////////////////////////////////////////
//
// TextFX8
// Copyright (c) 1995-2018 Jari Komppa
//
//
///////////////////////////////////////////////
// License
///////////////////////////////////////////////
// 
//     This software is provided 'as-is', without any express or implied
//     warranty.    In no event will the authors be held liable for any damages
//     arising from the use of this software.
// 
//     Permission is granted to anyone to use this software for any purpose,
//     including commercial applications, and to alter it and redistribute it
//     freely, subject to the following restrictions:
// 
//     1. The origin of this software must not be misrepresented; you must not
//        claim that you wrote the original software. If you use this software
//        in a product, an acknowledgment in the product documentation would be
//        appreciated but is not required.
//     2. Altered source versions must be plainly marked as such, and must not be
//        misrepresented as being the original software.
//     3. This notice may not be removed or altered from any source distribution.
// 
// (eg. same as ZLIB license)
// 
//
///////////////////////////////////////////////
#undef _CRT_SECURE_NO_WARNINGS
#define _CRT_SECURE_NO_WARNINGS
#include <stdlib.h>
#include <stdio.h>
#include "textfx.h"
#ifdef USE_STB_IMAGE
#include "stb_image.h"
#endif

unsigned short *lut_color2x2 = 0;

void alloc_lut_color2x2()
{
	lut_color2x2 = new unsigned short[64*64*64*64];
}

void load_lut_color2x2(char * aFilename)
{	
	if (lut_color2x2 == 0) alloc_lut_color2x2();
	FILE * f = fopen(aFilename, "rb");
	fread(lut_color2x2, 64*64*64*64*2, 1, f);
	fclose(f);
}

#ifdef USE_STB_IMAGE
void load_lut_color2x2_png(char * aFilename)
{
	alloc_lut_color2x2();
	int x, y, n;
    unsigned char *data = stbi_load(aFilename, &x, &y, &n, 0);
	for (x = 0; x < 64*64*64*64; x++)
	{
		((unsigned char*)lut_color2x2)[x*2] = data[x];
		((unsigned char*)lut_color2x2)[x*2+1] = data[x+64*64*64*64];
	}
    stbi_image_free(data);	
}
#endif

void convert_color2x2_160x100(unsigned int *fb, unsigned short *dst)
{
	int i, j;
	for (i = 0; i < 50; i++)
	{
		for (j = 0; j < 80; j++)
		{
			unsigned int px1 = fb[(i * 2 + 0) * 160 + (j * 2 + 0)];
			unsigned int px2 = fb[(i * 2 + 0) * 160 + (j * 2 + 1)];
			unsigned int px3 = fb[(i * 2 + 1) * 160 + (j * 2 + 0)];
			unsigned int px4 = fb[(i * 2 + 1) * 160 + (j * 2 + 1)];
			int r1 = ((px1 >>  0) & 0xff) >> 6;
			int g1 = ((px1 >>  8) & 0xff) >> 6;
			int b1 = ((px1 >> 16) & 0xff) >> 6;
			int r2 = ((px2 >>  0) & 0xff) >> 6;
			int g2 = ((px2 >>  8) & 0xff) >> 6;
			int b2 = ((px2 >> 16) & 0xff) >> 6;
			int r3 = ((px3 >>  0) & 0xff) >> 6;
			int g3 = ((px3 >>  8) & 0xff) >> 6;
			int b3 = ((px3 >> 16) & 0xff) >> 6;
			int r4 = ((px4 >>  0) & 0xff) >> 6;
			int g4 = ((px4 >>  8) & 0xff) >> 6;
			int b4 = ((px4 >> 16) & 0xff) >> 6;
			int index = 
				r1 +
				g1 * 4 +
				b1 * 4 * 4 +
				r2 * 4 * 4 * 4 +
				g2 * 4 * 4 * 4 * 4 +
				b2 * 4 * 4 * 4 * 4 * 4 +
				r3 * 4 * 4 * 4 * 4 * 4 * 4 +
				g3 * 4 * 4 * 4 * 4 * 4 * 4 * 4 +
				b3 * 4 * 4 * 4 * 4 * 4 * 4 * 4 * 4 +
				r4 * 4 * 4 * 4 * 4 * 4 * 4 * 4 * 4 * 4 +
				g4 * 4 * 4 * 4 * 4 * 4 * 4 * 4 * 4 * 4 * 4 +
				b4 * 4 * 4 * 4 * 4 * 4 * 4 * 4 * 4 * 4 * 4 * 4;
			dst[i * 80 + j] = lut_color2x2[index];
		}
	}
}

#ifdef TFX_SLOW_GENERATORS

float colors2x2[4*256]; // b/w "color" for the 4 segments of all chars

void extract_colors_2x2()
{
	int i, x, y, tx, ty;
	// Matrix to weight the samples of the font quadrant. Totally ad-hoc.
	float matrix[6*4] =
	{		
/*
		// solution
		1,1,1,1,
		1,1,1,1,
		1,1,1,1,
		1,1,1,1,
		1,1,1,1,
		1,1,1,1
*/
/*		
		// evolution
		2,4,4,2,
		4,5,5,4,
		5,7,7,5,
		5,7,7,5,
		4,5,5,4,
		2,4,4,2
*/
/*
		// revolution
		7,5,5,7,
		5,4,4,5,
		4,2,2,4,
		4,2,2,4,
		5,4,4,5,
		7,5,5,7
*/
/*
		// mutation
		1,2,3,4,
		3,4,1,2,
		1,2,3,4,
		3,4,1,2,
		1,2,3,4,
		3,4,1,2
*/
/*
		// evomutation
		2+1,4+2,4+1,2+2,
		4+3,5+4,5+3,4+4,
		5+1,7+2,7+1,5+2,
		5+3,7+4,7+3,5+4,
		4+1,5+2,5+1,4+2,
		2+3,4+4,4+3,2+4
*/
/*
		// desolution
		0,1,1,0,
		1,1,1,1,
		1,1,1,1,
		1,1,1,1,
		1,1,1,1,
		0,1,1,0
*/
/*
		// desolation
		0,1,1,0,
		1,0,0,1,
		1,0,0,1,
		1,0,0,1,
		1,0,0,1,
		0,1,1,0
*/
		// resolation
		1,2,2,1,
		2,1,1,2,
		2,1,1,2,
		2,1,1,2,
		2,1,1,2,
		1,2,2,1
	};

	float raw[8*12];
	int matrixsum = 0;
	for (i = 0; i < 6*4; i++)
		matrixsum += (int)matrix[i];

	for (i = 0; i < 256; i++)
	{
		for (y = 0; y < 12; y++)
		{
			for (x = 0; x < 8; x++)
			{
				raw[y * 8 + x] = IS_BIT(*(TFX_AsciiFontdata + i * 12 + y), x) ? 1.0f : 0.0f; 
			}
		}

		for (ty = 0; ty < 2; ty++)
		{
			for (tx = 0; tx < 2; tx++)
			{
				float sum = 0;
				for (y = 0; y < 6; y++)
				{
					for (x = 0; x < 4; x++)
					{
						sum += raw[(ty * 6 + y) * 8 + x + tx * 4] * matrix[y*6+x];
					}
				}
				colors2x2[i * 4 + ty * 2 + tx] = sum / (float)matrixsum;
			}
		}
	}
}

void build_lut_2x2()
{
	int r1, g1, b1, r2, g2, b2, r3, g3, b3, r4, g4, b4;
	float fr1, fg1, fb1, fr2, fg2, fb2, fr3, fg3, fb3, fr4, fg4, fb4;
	for (fr1 = 0, r1 = 0; r1 < 4; r1++, fr1 += 1.0f / 3.0f)
	{
		for (fr2 = 0, r2 = 0; r2 < 4; r2++, fr2 += 1.0f / 3.0f)
		{
			for (fr3 = 0, r3 = 0; r3 < 4; r3++, fr3 += 1.0f / 3.0f)
			{
				for (fg1 = 0, g1 = 0; g1 < 4; g1++, fg1 += 1.0f / 3.0f)
				{
					for (fg2 = 0, g2 = 0; g2 < 4; g2++, fg2 += 1.0f / 3.0f)
					{
						for (fg3 = 0, g3 = 0; g3 < 4; g3++, fg3 += 1.0f / 3.0f)
						{
							for (fb1 = 0, b1 = 0; b1 < 4; b1++, fb1 += 1.0f / 3.0f)
							{
								for (fb2 = 0, b2 = 0; b2 < 4; b2++, fb2 += 1.0f / 3.0f)
								{
									for (fb3 = 0, b3 = 0; b3 < 4; b3++, fb3 += 1.0f / 3.0f)
									{
										for (fr4 = 0, r4 = 0; r4 < 4; r4++, fr4 += 1.0f / 3.0f)
										{
											for (fg4 = 0, g4 = 0; g4 < 4; g4++, fg4 += 1.0f / 3.0f)
											{
												for (fb4 = 0, b4 = 0; b4 < 4; b4++, fb4 += 1.0f / 3.0f)
												{
													int index = 
														r1 +
														g1 * 4 +
														b1 * 4 * 4 +
														r2 * 4 * 4 * 4 +
														g2 * 4 * 4 * 4 * 4 +
														b2 * 4 * 4 * 4 * 4 * 4 +
														r3 * 4 * 4 * 4 * 4 * 4 * 4 +
														g3 * 4 * 4 * 4 * 4 * 4 * 4 * 4 +
														b3 * 4 * 4 * 4 * 4 * 4 * 4 * 4 * 4 +
														r4 * 4 * 4 * 4 * 4 * 4 * 4 * 4 * 4 * 4 +
														g4 * 4 * 4 * 4 * 4 * 4 * 4 * 4 * 4 * 4 * 4 +
														b4 * 4 * 4 * 4 * 4 * 4 * 4 * 4 * 4 * 4 * 4 * 4;
													int fg, bg, ch;
													int closest = 0;
													float distance = 1e8;
													for (fg = 0; fg < 16; fg++)
													{
														for (bg = 0; bg < 16; bg++)
														{
															for (ch = 0; ch < 256; ch++)
															{
																float cr1 = (colors2x2[ch * 4 + 0] * TFX_Palette8[fg * 3 + 2] + (1 - colors2x2[ch * 4 + 0]) * TFX_Palette8[bg * 3 + 2]) / 256.0f;
																float cg1 = (colors2x2[ch * 4 + 0] * TFX_Palette8[fg * 3 + 1] + (1 - colors2x2[ch * 4 + 0]) * TFX_Palette8[bg * 3 + 1]) / 256.0f;
																float cb1 = (colors2x2[ch * 4 + 0] * TFX_Palette8[fg * 3 + 0] + (1 - colors2x2[ch * 4 + 0]) * TFX_Palette8[bg * 3 + 0]) / 256.0f;

																float cr2 = (colors2x2[ch * 4 + 1] * TFX_Palette8[fg * 3 + 2] + (1 - colors2x2[ch * 4 + 1]) * TFX_Palette8[bg * 3 + 2]) / 256.0f;
																float cg2 = (colors2x2[ch * 4 + 1] * TFX_Palette8[fg * 3 + 1] + (1 - colors2x2[ch * 4 + 1]) * TFX_Palette8[bg * 3 + 1]) / 256.0f;
																float cb2 = (colors2x2[ch * 4 + 1] * TFX_Palette8[fg * 3 + 0] + (1 - colors2x2[ch * 4 + 1]) * TFX_Palette8[bg * 3 + 0]) / 256.0f;

																float cr3 = (colors2x2[ch * 4 + 2] * TFX_Palette8[fg * 3 + 2] + (1 - colors2x2[ch * 4 + 2]) * TFX_Palette8[bg * 3 + 2]) / 256.0f;
																float cg3 = (colors2x2[ch * 4 + 2] * TFX_Palette8[fg * 3 + 1] + (1 - colors2x2[ch * 4 + 2]) * TFX_Palette8[bg * 3 + 1]) / 256.0f;
																float cb3 = (colors2x2[ch * 4 + 2] * TFX_Palette8[fg * 3 + 0] + (1 - colors2x2[ch * 4 + 2]) * TFX_Palette8[bg * 3 + 0]) / 256.0f;

																float cr4 = (colors2x2[ch * 4 + 3] * TFX_Palette8[fg * 3 + 2] + (1 - colors2x2[ch * 4 + 3]) * TFX_Palette8[bg * 3 + 2]) / 256.0f;
																float cg4 = (colors2x2[ch * 4 + 3] * TFX_Palette8[fg * 3 + 1] + (1 - colors2x2[ch * 4 + 3]) * TFX_Palette8[bg * 3 + 1]) / 256.0f;
																float cb4 = (colors2x2[ch * 4 + 3] * TFX_Palette8[fg * 3 + 0] + (1 - colors2x2[ch * 4 + 3]) * TFX_Palette8[bg * 3 + 0]) / 256.0f;

																float candidate_distance = 
																	SQUARE(fr1-cr1) + SQUARE(fg1-cg1) + SQUARE(fb1-cb1) +
																	SQUARE(fr2-cr2) + SQUARE(fg2-cg2) + SQUARE(fb2-cb2) +
																	SQUARE(fr3-cr3) + SQUARE(fg3-cg3) + SQUARE(fb3-cb3) +
																	SQUARE(fr4-cr4) + SQUARE(fg4-cg4) + SQUARE(fb4-cb4);
																if (distance > candidate_distance)
																{
																	distance = candidate_distance;
																	closest = ch | (fg << 8) | (bg << 12);
																}
															}
														}
														lut_color2x2[index] = closest;
													}		
												}
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}
	}
}
#endif