///////////////////////////////////////////////
// Copyright
///////////////////////////////////////////////
//
// TextFX8
// Copyright (c) 1995-2018 Jari Komppa
//
//
///////////////////////////////////////////////
// License
///////////////////////////////////////////////
// 
//     This software is provided 'as-is', without any express or implied
//     warranty.    In no event will the authors be held liable for any damages
//     arising from the use of this software.
// 
//     Permission is granted to anyone to use this software for any purpose,
//     including commercial applications, and to alter it and redistribute it
//     freely, subject to the following restrictions:
// 
//     1. The origin of this software must not be misrepresented; you must not
//        claim that you wrote the original software. If you use this software
//        in a product, an acknowledgment in the product documentation would be
//        appreciated but is not required.
//     2. Altered source versions must be plainly marked as such, and must not be
//        misrepresented as being the original software.
//     3. This notice may not be removed or altered from any source distribution.
// 
// (eg. same as ZLIB license)
// 
//
///////////////////////////////////////////////

#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include "textfx.h"


static int findcol(int incol)
{
    int dist = 0xffffff;
    int col = 0;
    int i;
    for (i = 0; i < 16; i++)
    {
        int d = SQUARE(TFX_Palette8[i * 3 + 2]-((incol >> 0) & 0xff)) +
                SQUARE(TFX_Palette8[i * 3 + 1]-((incol >> 8) & 0xff)) +
                SQUARE(TFX_Palette8[i * 3 + 0]-((incol >> 16) & 0xff));

        if (d < dist)
        {
            col = i;
            dist = d;
        }
    }

    return col;
}


void convert_halfblockcolor_160x100(unsigned int *aSource, unsigned short *aTarget)
{
	int x, y, c;
    for (y = 0, c = 0; y < 50; y++)
    {
        for (x = 0; x < 80; x++, c++)
        {
            unsigned int c1 = aSource[(y*2+0)*160+x*2+0];
            unsigned int c2 = aSource[(y*2+0)*160+x*2+1];
            unsigned int c3 = aSource[(y*2+1)*160+x*2+0];
            unsigned int c4 = aSource[(y*2+1)*160+x*2+1];


            unsigned int redblue;
            unsigned int green;
            redblue  = c1 & 0xff00ff;
            green    = c1 & 0x00ff00;
            redblue += c3 & 0xff00ff;
            green   += c3 & 0x00ff00;
            int color1w = ((redblue>>1) & 0xff00ff) + ((green>>1) & 0x00ff00);
            redblue = green = 0;
            redblue += c2 & 0xff00ff;
            green   += c2 & 0x00ff00;
            redblue += c4 & 0xff00ff;
            green   += c4 & 0x00ff00;
            int color2w = ((redblue>>1) & 0xff00ff) + ((green>>1) & 0x00ff00);
            redblue  = c1 & 0xff00ff;
            green    = c1 & 0x00ff00;
            redblue += c2 & 0xff00ff;
            green   += c2 & 0x00ff00;
            int color1h = ((redblue>>1) & 0xff00ff) + ((green>>1) & 0x00ff00);
            redblue = green = 0;
            redblue += c3 & 0xff00ff;
            green   += c3 & 0x00ff00;
            redblue += c4 & 0xff00ff;
            green   += c4 & 0x00ff00;
            int color2h = ((redblue>>1) & 0xff00ff) + ((green>>1) & 0x00ff00);

            int dw = SQUARE(((color1w >> 0) & 0xff)-((color2w >> 0) & 0xff)) +
                SQUARE(((color1w >> 8) & 0xff)-((color2w >> 8) & 0xff)) +
                SQUARE(((color1w >> 16) & 0xff)-((color2w >> 16) & 0xff));

            int dh = SQUARE(((color1h >> 0) & 0xff)-((color2h >> 0) & 0xff)) +
                SQUARE(((color1h >> 8) & 0xff)-((color2h >> 8) & 0xff)) +
                SQUARE(((color1h >> 16) & 0xff)-((color2h >> 16) & 0xff));

            if (dw < dh)
            {
                int col1 = findcol(color2h);
                int col2 = findcol(color1h);

                aTarget[c] = 220 | (col1 << 8) | (col2 << 12);

            }
            else
            {
                int col1 = findcol(color2w);
                int col2 = findcol(color1w);

                aTarget[c] = 221 | (col2 << 8) | (col1 << 12);
            }
        }
    }
}

