///////////////////////////////////////////////
// Copyright
///////////////////////////////////////////////
//
// TextFX8
// Copyright (c) 1995-2018 Jari Komppa
//
//
///////////////////////////////////////////////
// License
///////////////////////////////////////////////
// 
//     This software is provided 'as-is', without any express or implied
//     warranty.    In no event will the authors be held liable for any damages
//     arising from the use of this software.
// 
//     Permission is granted to anyone to use this software for any purpose,
//     including commercial applications, and to alter it and redistribute it
//     freely, subject to the following restrictions:
// 
//     1. The origin of this software must not be misrepresented; you must not
//        claim that you wrote the original software. If you use this software
//        in a product, an acknowledgment in the product documentation would be
//        appreciated but is not required.
//     2. Altered source versions must be plainly marked as such, and must not be
//        misrepresented as being the original software.
//     3. This notice may not be removed or altered from any source distribution.
// 
// (eg. same as ZLIB license)
// 
//
///////////////////////////////////////////////

#undef _CRT_SECURE_NO_WARNINGS
#define _CRT_SECURE_NO_WARNINGS
#include <stdlib.h>
#include <stdio.h>
#include "textfx.h"
#ifdef USE_STB_IMAGE
#include "stb_image.h"
#endif

unsigned short *lut_hires = 0;

void alloc_lut_hires()
{
	lut_hires = new unsigned short[3*3*3*3* 3*3*3*3* 3*3*3*3* 3*3*3*3];
}

void load_lut_hires(char * aFilename)
{	
	if (lut_hires == 0) alloc_lut_hires();
	FILE * f = fopen(aFilename, "rb");
	fread(lut_hires, 3*3*3*3* 3*3*3*3* 3*3*3*3* 3*3*3*3 *2, 1, f);
	fclose(f);
}


#ifdef USE_STB_IMAGE
void load_lut_hires_png(char * aFilename)
{
	alloc_lut_hires();
	int x, y, n;
    unsigned char *data = stbi_load(aFilename, &x, &y, &n, 0);
	for (x = 0; x < 3*3*3*3* 3*3*3*3* 3*3*3*3* 3*3*3*3; x++)
	{
		((unsigned char*)lut_hires)[x*2] = data[x];
		((unsigned char*)lut_hires)[x*2+1] = data[x+3*3*3*3* 3*3*3*3* 3*3*3*3* 3*3*3*3];
	}
    stbi_image_free(data);	
}
#endif


void convert_hires_320x200(unsigned int *fb, unsigned short *dst)
{
	int i, j;
	for (i = 0; i < 50; i++)
	{
		for (j = 0; j < 80; j++)
		{
			unsigned int px11 = fb[(i * 4 + 0) * 320 + (j * 4 + 0)];
			unsigned int px12 = fb[(i * 4 + 0) * 320 + (j * 4 + 1)];
			unsigned int px13 = fb[(i * 4 + 0) * 320 + (j * 4 + 2)];
			unsigned int px14 = fb[(i * 4 + 0) * 320 + (j * 4 + 3)];
			unsigned int px21 = fb[(i * 4 + 1) * 320 + (j * 4 + 0)];
			unsigned int px22 = fb[(i * 4 + 1) * 320 + (j * 4 + 1)];
			unsigned int px23 = fb[(i * 4 + 1) * 320 + (j * 4 + 2)];
			unsigned int px24 = fb[(i * 4 + 1) * 320 + (j * 4 + 3)];
			unsigned int px31 = fb[(i * 4 + 2) * 320 + (j * 4 + 0)];
			unsigned int px32 = fb[(i * 4 + 2) * 320 + (j * 4 + 1)];
			unsigned int px33 = fb[(i * 4 + 2) * 320 + (j * 4 + 2)];
			unsigned int px34 = fb[(i * 4 + 2) * 320 + (j * 4 + 3)];
			unsigned int px41 = fb[(i * 4 + 3) * 320 + (j * 4 + 0)];
			unsigned int px42 = fb[(i * 4 + 3) * 320 + (j * 4 + 1)];
			unsigned int px43 = fb[(i * 4 + 3) * 320 + (j * 4 + 2)];
			unsigned int px44 = fb[(i * 4 + 3) * 320 + (j * 4 + 3)];

#define GS(x) int i##x = ((((px##x >> 0) & 0xff) * 2 + ((px##x >> 8) & 0xff) * 5 + ((px##x >> 16) & 0xff)) * 3) >> (8+3);

			GS(11) GS(12) GS(13) GS(14)
			GS(21) GS(22) GS(23) GS(24)
			GS(31) GS(32) GS(33) GS(34)
			GS(41) GS(42) GS(43) GS(44)			

#undef GS

			int index =
				i11 +
				i12 * 3 +
				i13 * 3 * 3 +
				i14 * 3 * 3 * 3 +
				i21 * 3 * 3 * 3 * 3 +
				i22 * 3 * 3 * 3 * 3 * 3 +
				i23 * 3 * 3 * 3 * 3 * 3 * 3 +
				i24 * 3 * 3 * 3 * 3 * 3 * 3 * 3 +
				i31 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 +
				i32 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 +
				i33 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 +
				i34 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 +
				i41 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 +
				i42 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 +
				i43 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 +
				i44 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3;
			TFX_FrameBuffer[i * 80 + j] = lut_hires[index];
		}
	}
}

#ifdef TFX_SLOW_GENERATORS
float colors4x4[16*256]; // b/w "color" for the 16 segments of all chars

void extract_colors_4x4()
{
	int i, x, y, tx, ty;
	float raw[8*12];

	for (i = 0; i < 256; i++)
	{
		for (y = 0; y < 12; y++)
		{
			for (x = 0; x < 8; x++)
			{
				raw[y * 8 + x] = IS_BIT(*(TFX_AsciiFontdata + i * 12 + y), x) ? 1.0f : 0.0f; 
			}
		}

		for (ty = 0; ty < 4; ty++)
		{
			for (tx = 0; tx < 4; tx++)
			{
				float sum = 0;
				for (y = 0; y < 3; y++)
				{
					for (x = 0; x < 2; x++)
					{
						sum += raw[(ty * 3 + y) * 8 + x + tx * 2];
					}
				}
				colors4x4[i * 16 + ty * 4 + tx] = sum / (float)(3*2);
			}
		}
	}
}

void build_lut_4x4()
{
	int coloridx[3] = { 0, 7, 15 };
	int i11, i12, i13, i14,
		i21, i22, i23, i24,
		i31, i32, i33, i34,
		i41, i42, i43, i44;
	float f[16];
	
#define FORX(x,y) for (i##x = 0, f[y] = 0; i##x < 3; i##x++, f[y] += 0.5f) {
	FORX(11,0) FORX(12,1) FORX(13,2) FORX(14,3)
	FORX(21,4) FORX(22,5) FORX(23,6) FORX(24,7)
	FORX(31,8) FORX(32,9) FORX(33,10) FORX(34,11)
	FORX(41,12) FORX(42,13) FORX(43,14) FORX(44,15)
		int index =
			i11 +
			i12 * 3 +
			i13 * 3 * 3 +
			i14 * 3 * 3 * 3 +
			i21 * 3 * 3 * 3 * 3 +
			i22 * 3 * 3 * 3 * 3 * 3 +
			i23 * 3 * 3 * 3 * 3 * 3 * 3 +
			i24 * 3 * 3 * 3 * 3 * 3 * 3 * 3 +
			i31 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 +
			i32 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 +
			i33 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 +
			i34 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 +
			i41 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 +
			i42 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 +
			i43 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 +
			i44 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3;
		int fg, bg, ch, qr;
		int closest = 0;
		float distance = 1e8;
		for (fg = 0; fg < 3; fg++)
		{
			for (bg = 0; bg < 3; bg++)
			{				
				for (ch = 0; ch < 256; ch++)
				{
					float candidate_distance = 0;
					for (qr = 0; qr < 16; qr++)
					{
						float cc = (colors4x4[ch * 16 + qr] * TFX_Palette8[coloridx[fg] * 3 + 2] + (1 - colors4x4[ch * 16 + qr]) * TFX_Palette8[coloridx[bg] * 3 + 2]) / 256.0f;
						candidate_distance += SQUARE(f[qr]-cc);
					}
					if (distance > candidate_distance)
					{
						distance = candidate_distance;
						closest = ch | (coloridx[fg] << 8) | (coloridx[bg] << 12);
					}
				}
			}
		}
		lut_hires[index] = closest;
	} } } }
	} } } }
	} } } }
	} } } }
}

#endif